"""
Original author: Joshua Haberman
https://github.com/haberman/vtparse/blob/198ea4382f824dbb3f0e5b5553a9eb3290764694/vtparse_tables.rb

==

https://vt100.net/emu/dec_ansi_parser
https://github.com/haberman/vtparse
"""
import typing as ta


##


class StateTransition(ta.NamedTuple):
    to_state: str


TransitionMap: ta.TypeAlias = ta.Mapping[
    int | range | str,
    str | StateTransition | tuple[str | StateTransition, ...],
]


##


# Define the anywhere transitions
ANYWHERE_TRANSITIONS: TransitionMap = {
    0x18: ('execute', StateTransition('ground')),
    0x1a: ('execute', StateTransition('ground')),
    range(0x80, 0x90): ('execute', StateTransition('ground')),
    range(0x91, 0x98): ('execute', StateTransition('ground')),
    0x99: ('execute', StateTransition('ground')),
    0x9a: ('execute', StateTransition('ground')),
    0x9c: StateTransition('ground'),
    0x1b: StateTransition('escape'),
    0x98: StateTransition('sos_pm_apc_string'),
    0x9e: StateTransition('sos_pm_apc_string'),
    0x9f: StateTransition('sos_pm_apc_string'),
    0x90: StateTransition('dcs_entry'),
    0x9d: StateTransition('osc_string'),
    0x9b: StateTransition('csi_entry'),
}


# Global states dictionary
STATES: ta.Mapping[str, TransitionMap] = {
    # Define state transitions
    'ground': {
        range(0x18): 'execute',
        0x19: 'execute',
        range(0x1c, 0x20): 'execute',
        range(0x20, 0x80): 'print',
    },

    'escape': {
        'on_entry': 'clear',
        range(0x18): 'execute',
        0x19: 'execute',
        range(0x1c, 0x20): 'execute',
        0x7f: 'ignore',
        range(0x20, 0x30): ('collect', StateTransition('escape_intermediate')),
        range(0x30, 0x50): ('esc_dispatch', StateTransition('ground')),
        range(0x51, 0x58): ('esc_dispatch', StateTransition('ground')),
        0x59: ('esc_dispatch', StateTransition('ground')),
        0x5a: ('esc_dispatch', StateTransition('ground')),
        0x5c: ('esc_dispatch', StateTransition('ground')),
        range(0x60, 0x7f): ('esc_dispatch', StateTransition('ground')),
        0x5b: StateTransition('csi_entry'),
        0x5d: StateTransition('osc_string'),
        0x50: StateTransition('dcs_entry'),
        0x58: StateTransition('sos_pm_apc_string'),
        0x5e: StateTransition('sos_pm_apc_string'),
        0x5f: StateTransition('sos_pm_apc_string'),
    },

    'escape_intermediate': {
        range(0x18): 'execute',
        0x19: 'execute',
        range(0x1c, 0x20): 'execute',
        range(0x20, 0x30): 'collect',
        0x7f: 'ignore',
        range(0x30, 0x7f): ('esc_dispatch', StateTransition('ground')),
    },

    'csi_entry': {
        'on_entry': 'clear',
        range(0x18): 'execute',
        0x19: 'execute',
        range(0x1c, 0x20): 'execute',
        0x7f: 'ignore',
        range(0x20, 0x30): ('collect', StateTransition('csi_intermediate')),
        0x3a: StateTransition('csi_ignore'),
        range(0x30, 0x3a): ('param', StateTransition('csi_param')),
        0x3b: ('param', StateTransition('csi_param')),
        range(0x3c, 0x40): ('collect', StateTransition('csi_param')),
        range(0x40, 0x7f): ('csi_dispatch', StateTransition('ground')),
    },

    'csi_ignore': {
        range(0x18): 'execute',
        0x19: 'execute',
        range(0x1c, 0x20): 'execute',
        range(0x20, 0x40): 'ignore',
        0x7f: 'ignore',
        range(0x40, 0x7f): StateTransition('ground'),
    },

    'csi_param': {
        range(0x18): 'execute',
        0x19: 'execute',
        range(0x1c, 0x20): 'execute',
        range(0x30, 0x3a): 'param',
        0x3b: 'param',
        0x7f: 'ignore',
        0x3a: StateTransition('csi_ignore'),
        range(0x3c, 0x40): StateTransition('csi_ignore'),
        range(0x20, 0x30): ('collect', StateTransition('csi_intermediate')),
        range(0x40, 0x7f): ('csi_dispatch', StateTransition('ground')),
    },

    'csi_intermediate': {
        range(0x18): 'execute',
        0x19: 'execute',
        range(0x1c, 0x20): 'execute',
        range(0x20, 0x30): 'collect',
        0x7f: 'ignore',
        range(0x30, 0x40): StateTransition('csi_ignore'),
        range(0x40, 0x7f): ('csi_dispatch', StateTransition('ground')),
    },

    'dcs_entry': {
        'on_entry': 'clear',
        range(0x18): 'ignore',
        0x19: 'ignore',
        range(0x1c, 0x20): 'ignore',
        0x7f: 'ignore',
        0x3a: StateTransition('dcs_ignore'),
        range(0x20, 0x30): ('collect', StateTransition('dcs_intermediate')),
        range(0x30, 0x3a): ('param', StateTransition('dcs_param')),
        0x3b: ('param', StateTransition('dcs_param')),
        range(0x3c, 0x40): ('collect', StateTransition('dcs_param')),
        range(0x40, 0x7f): (StateTransition('dcs_passthrough')),
    },

    'dcs_intermediate': {
        range(0x18): 'ignore',
        0x19: 'ignore',
        range(0x1c, 0x20): 'ignore',
        range(0x20, 0x30): 'collect',
        0x7f: 'ignore',
        range(0x30, 0x40): StateTransition('dcs_ignore'),
        range(0x40, 0x7f): StateTransition('dcs_passthrough'),
    },

    'dcs_ignore': {
        range(0x18): 'ignore',
        0x19: 'ignore',
        range(0x1c, 0x80): 'ignore',
    },

    'dcs_param': {
        range(0x18): 'ignore',
        0x19: 'ignore',
        range(0x1c, 0x20): 'ignore',
        range(0x30, 0x3a): 'param',
        0x3b: 'param',
        0x7f: 'ignore',
        0x3a: StateTransition('dcs_ignore'),
        range(0x3c, 0x40): StateTransition('dcs_ignore'),
        range(0x20, 0x30): ('collect', StateTransition('dcs_intermediate')),
        range(0x40, 0x7f): StateTransition('dcs_passthrough'),
    },

    'dcs_passthrough': {
        'on_entry': 'hook',
        range(0x18): 'put',
        0x19: 'put',
        range(0x1c, 0x20): 'put',
        range(0x20, 0x7f): 'put',
        0x7f: 'ignore',
        'on_exit': 'unhook',
    },

    'sos_pm_apc_string': {
        range(0x18): 'ignore',
        0x19: 'ignore',
        range(0x1c, 0x80): 'ignore',
    },

    'osc_string': {
        'on_entry': 'osc_start',
        range(0x18): 'ignore',
        0x19: 'ignore',
        range(0x1c, 0x20): 'ignore',
        range(0x20, 0x80): 'osc_put',
        'on_exit': 'osc_end',
    },
}


##


ACTIONS_IN_ORDER = sorted(
    action
    for state, transitions in STATES.items()
    for keys, actions in transitions.items()
    for action in ([actions] if not isinstance(actions, tuple) else actions)
    if isinstance(action, str)
)

STATES_IN_ORDER = sorted(STATES)


##


TransitionTable: ta.TypeAlias = tuple[tuple[str | StateTransition, ...], ...]


def _build_state_tables() -> ta.Mapping[str, TransitionTable]:
    # Expand the range-based data structures into fully expanded tables
    state_tables: dict[str, list] = {}

    def expand_ranges(dct):
        array = [None] * 256
        for k, v in dct.items():
            if isinstance(k, range):
                for i in k:
                    array[i] = v
            elif isinstance(k, int):
                array[k] = v
        return array

    for s, t in STATES.items():
        state_tables[s] = expand_ranges(t)

    # Seed all the states with the anywhere transitions
    anywhere_transitions_expanded = expand_ranges(ANYWHERE_TRANSITIONS)

    for state, transitions in state_tables.items():
        for i, transition in enumerate(anywhere_transitions_expanded):
            if transition is not None:
                if transitions[i] is not None:
                    raise ValueError(
                        f'State {state} already had a transition defined for 0x{i:02x}, but that transition is also an '
                        f'anywhere transition!',
                    )
                transitions[i] = transition

    # For consistency, make all transitions tuples of actions
    return {
        state: tuple(tuple(t if isinstance(t, (list, tuple)) else [t]) for t in transitions)
        for state, transitions in state_tables.items()
    }


STATE_TABLES = _build_state_tables()


##


def check_state_tables(state_tables: ta.Mapping[str, TransitionTable]) -> None:
    for state, transitions in state_tables.items():
        for i, val in enumerate(transitions):
            if not val:
                raise ValueError(f'No transition defined from state {state}, char 0x{i:02x}!')



# ruff: noqa: UP006 UP007 UP043 UP045
# @omlish-lite
##
# MIT License
#
# Copyright (c) 2019 Masaaki Goshima
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##
import copy
import dataclasses as dc
import datetime
import enum
import functools
import typing as ta

from ....lite.dataclasses import dataclass_field_required
from .errors import YamlError
from .errors import YamlErrorOr
from .errors import yaml_error


##


@dc.dataclass(frozen=True)
class YamlSyntaxError(YamlError):
    msg: str
    token: ta.Optional['YamlToken']

    @property
    def message(self) -> str:
        return self.msg


##


class YamlChars:
    # SEQUENCE_ENTRY character for sequence entry
    SEQUENCE_ENTRY = '-'
    # MAPPING_KEY character for mapping key
    MAPPING_KEY = '?'
    # MAPPING_VALUE character for mapping value
    MAPPING_VALUE = ':'
    # COLLECT_ENTRY character for collect entry
    COLLECT_ENTRY = ','
    # SEQUENCE_START character for sequence start
    SEQUENCE_START = '['
    # SEQUENCE_END character for sequence end
    SEQUENCE_END = ']'
    # MAPPING_START character for mapping start
    MAPPING_START = '{'
    # MAPPING_END character for mapping end
    MAPPING_END = '}'
    # COMMENT character for comment
    COMMENT = '#'
    # ANCHOR character for anchor
    ANCHOR = '&'
    # ALIAS character for alias
    ALIAS = '*'
    # TAG character for tag
    TAG = '!'
    # LITERAL character for literal
    LITERAL = '|'
    # FOLDED character for folded
    FOLDED = '>'
    # SINGLE_QUOTE character for single quote
    SINGLE_QUOTE = '\''
    # DOUBLE_QUOTE character for double quote
    DOUBLE_QUOTE = '"'
    # DIRECTIVE character for directive
    DIRECTIVE = '%'
    # SPACE character for space
    SPACE = ' '
    # LINE_BREAK character for line break
    LINE_BREAK = '\n'


class YamlTokenType(enum.Enum):
    # UNKNOWN reserve for invalid type
    UNKNOWN = enum.auto()
    # DOCUMENT_HEADER type for DocumentHeader token
    DOCUMENT_HEADER = enum.auto()
    # DOCUMENT_END type for DocumentEnd token
    DOCUMENT_END = enum.auto()
    # SEQUENCE_ENTRY type for SequenceEntry token
    SEQUENCE_ENTRY = enum.auto()
    # MAPPING_KEY type for MappingKey token
    MAPPING_KEY = enum.auto()
    # MAPPING_VALUE type for MappingValue token
    MAPPING_VALUE = enum.auto()
    # MERGE_KEY type for MergeKey token
    MERGE_KEY = enum.auto()
    # COLLECT_ENTRY type for CollectEntry token
    COLLECT_ENTRY = enum.auto()
    # SEQUENCE_START type for SequenceStart token
    SEQUENCE_START = enum.auto()
    # SEQUENCE_END type for SequenceEnd token
    SEQUENCE_END = enum.auto()
    # MAPPING_START type for MappingStart token
    MAPPING_START = enum.auto()
    # MAPPING_END type for MappingEnd token
    MAPPING_END = enum.auto()
    # COMMENT type for Comment token
    COMMENT = enum.auto()
    # ANCHOR type for Anchor token
    ANCHOR = enum.auto()
    # ALIAS type for Alias token
    ALIAS = enum.auto()
    # TAG type for Tag token
    TAG = enum.auto()
    # LITERAL type for Literal token
    LITERAL = enum.auto()
    # FOLDED type for Folded token
    FOLDED = enum.auto()
    # SINGLE_QUOTE type for SingleQuote token
    SINGLE_QUOTE = enum.auto()
    # DOUBLE_QUOTE type for DoubleQuote token
    DOUBLE_QUOTE = enum.auto()
    # DIRECTIVE type for Directive token
    DIRECTIVE = enum.auto()
    # SPACE type for Space token
    SPACE = enum.auto()
    # NULL type for Null token
    NULL = enum.auto()
    # IMPLICIT_NULL type for implicit Null token.
    # This is used when explicit keywords such as null or ~ are not specified. It is distinguished during encoding and
    # output as an empty string.
    IMPLICIT_NULL = enum.auto()
    # INFINITY type for Infinity token
    INFINITY = enum.auto()
    # NAN type for Nan token
    NAN = enum.auto()
    # INTEGER type for Integer token
    INTEGER = enum.auto()
    # BINARY_INTEGER type for BinaryInteger token
    BINARY_INTEGER = enum.auto()
    # OCTET_INTEGER type for OctetInteger token
    OCTET_INTEGER = enum.auto()
    # HEX_INTEGER type for HexInteger token
    HEX_INTEGER = enum.auto()
    # FLOAT type for Float token
    FLOAT = enum.auto()
    # STRING type for String token
    STRING = enum.auto()
    # BOOL type for Bool token
    BOOL = enum.auto()
    # INVALID type for invalid token
    INVALID = enum.auto()


class YamlCharType(enum.Enum):
    # INDICATOR type of indicator character
    INDICATOR = enum.auto()
    # WHITE-SPACE type of white space character
    WHITESPACE = enum.auto()
    # MISCELLANEOUS type of miscellaneous character
    MISCELLANEOUS = enum.auto()
    # ESCAPED type of escaped character
    ESCAPED = enum.auto()
    # INVALID type for an invalid token.
    INVALID = enum.auto()


class YamlIndicator(enum.Enum):
    # NOT not an indicator
    NOT = enum.auto()
    # BLOCK_STRUCTURE indicator for block structure ( '-', '?', ':' )
    BLOCK_STRUCTURE = enum.auto()
    # FLOW_COLLECTION indicator for flow collection ( '[', ']', '{', '}', ',' )
    FLOW_COLLECTION = enum.auto()
    # COMMENT indicator for comment ( '#' )
    COMMENT = enum.auto()
    # NODE_PROPERTY indicator for node property ( '!', '&', '*' )
    NODE_PROPERTY = enum.auto()
    # BLOCK_SCALAR indicator for block scalar ( '|', '>' )
    BLOCK_SCALAR = enum.auto()
    # QUOTED_SCALAR indicator for quoted scalar ( ''', '"' )
    QUOTED_SCALAR = enum.auto()
    # DIRECTIVE indicator for directive ( '%' )
    DIRECTIVE = enum.auto()
    # INVALID_USE_OF_RESERVED indicator for invalid use of reserved keyword ( '@', '`' )
    INVALID_USE_OF_RESERVED = enum.auto()


##


class YamlKeywords:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    RESERVED_NULL_KEYWORDS = (
        'null',
        'Null',
        'NULL',
        '~',
    )

    RESERVED_BOOL_KEYWORDS = (
        'true',
        'True',
        'TRUE',
        'false',
        'False',
        'FALSE',
    )

    # For compatibility with other YAML 1.1 parsers.
    # Note that we use these solely for encoding the bool value with quotes. go-yaml should not treat these as reserved
    # keywords at parsing time. as go-yaml is supposed to be compliant only with YAML 1.2.
    RESERVED_LEGACY_BOOL_KEYWORDS = (
        'y',
        'Y',
        'yes',
        'Yes',
        'YES',
        'n',
        'N',
        'no',
        'No',
        'NO',
        'on',
        'On',
        'ON',
        'off',
        'Off',
        'OFF',
    )

    RESERVED_INF_KEYWORDS = (
        '.inf',
        '.Inf',
        '.INF',
        '-.inf',
        '-.Inf',
        '-.INF',
    )

    RESERVED_NAN_KEYWORDS = (
        '.nan',
        '.NaN',
        '.NAN',
    )

    RESERVED_KEYWORD_MAP: ta.ClassVar[ta.Mapping[str, ta.Callable[[str, str, 'YamlPosition'], 'YamlToken']]]
    RESERVED_ENC_KEYWORD_MAP: ta.ClassVar[ta.Mapping[str, ta.Callable[[str, str, 'YamlPosition'], 'YamlToken']]]


def _yaml_reserved_keyword_token(typ: YamlTokenType, value: str, org: str, pos: 'YamlPosition') -> 'YamlToken':
    return YamlToken(
        type=typ,
        char_type=YamlCharType.MISCELLANEOUS,
        indicator=YamlIndicator.NOT,
        value=value,
        origin=org,
        position=pos,
    )


YamlKeywords.RESERVED_KEYWORD_MAP = {
    **{keyword: functools.partial(_yaml_reserved_keyword_token, YamlTokenType.NULL) for keyword in YamlKeywords.RESERVED_NULL_KEYWORDS},  # noqa
    **{keyword: functools.partial(_yaml_reserved_keyword_token, YamlTokenType.BOOL) for keyword in YamlKeywords.RESERVED_BOOL_KEYWORDS},  # noqa
    **{keyword: functools.partial(_yaml_reserved_keyword_token, YamlTokenType.INFINITY) for keyword in YamlKeywords.RESERVED_INF_KEYWORDS},  # noqa
    **{keyword: functools.partial(_yaml_reserved_keyword_token, YamlTokenType.NAN) for keyword in YamlKeywords.RESERVED_NAN_KEYWORDS},  # noqa
}


# RESERVED_ENC_KEYWORD_MAP contains is the keyword map used at encoding time.
# This is supposed to be a superset of RESERVED_KEYWORD_MAP, and used to quote legacy keywords present in YAML 1.1 or
# lesser for compatibility reasons, even though this library is supposed to be YAML 1.2-compliant.
YamlKeywords.RESERVED_ENC_KEYWORD_MAP = {
    **{keyword: functools.partial(_yaml_reserved_keyword_token, YamlTokenType.NULL) for keyword in YamlKeywords.RESERVED_NULL_KEYWORDS},  # noqa
    **{keyword: functools.partial(_yaml_reserved_keyword_token, YamlTokenType.BOOL) for keyword in YamlKeywords.RESERVED_BOOL_KEYWORDS},  # noqa
    **{keyword: functools.partial(_yaml_reserved_keyword_token, YamlTokenType.BOOL) for keyword in YamlKeywords.RESERVED_LEGACY_BOOL_KEYWORDS},  # noqa
}


##


YamlReservedTagKeyword = str  # ta.TypeAlias  # omlish-amalg-typing-no-move


class YamlReservedTagKeywords:
    # INTEGER `!!int` tag
    INTEGER = '!!int'
    # FLOAT `!!float` tag
    FLOAT = '!!float'
    # NULL `!!null` tag
    NULL = '!!null'
    # SEQUENCE `!!seq` tag
    SEQUENCE = '!!seq'
    # MAPPING `!!map` tag
    MAPPING = '!!map'
    # STRING `!!str` tag
    STRING = '!!str'
    # BINARY `!!binary` tag
    BINARY = '!!binary'
    # ORDERED_MAP `!!omap` tag
    ORDERED_MAP = '!!omap'
    # SET `!!set` tag
    SET = '!!set'
    # TIMESTAMP `!!timestamp` tag
    TIMESTAMP = '!!timestamp'
    # BOOLEAN `!!bool` tag
    BOOLEAN = '!!bool'
    # MERGE `!!merge` tag
    MERGE = '!!merge'


# RESERVED_TAG_KEYWORD_MAP map for reserved tag keywords
YAML_RESERVED_TAG_KEYWORD_MAP: ta.Mapping[YamlReservedTagKeyword, ta.Callable[[str, str, 'YamlPosition'], 'YamlToken']] = {  # noqa
    YamlReservedTagKeywords.INTEGER: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.FLOAT: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.NULL: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.SEQUENCE: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.MAPPING: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.STRING: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.BINARY: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.ORDERED_MAP: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.SET: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.TIMESTAMP: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.BOOLEAN: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
    YamlReservedTagKeywords.MERGE: lambda value, org, pos: YamlToken(
        type=YamlTokenType.TAG,
        char_type=YamlCharType.INDICATOR,
        indicator=YamlIndicator.NODE_PROPERTY,
        value=value,
        origin=org,
        position=pos,
    ),
}


##


class YamlNumberType(enum.Enum):
    DECIMAL = 'decimal'
    BINARY = 'binary'
    OCTET = 'octet'
    HEX = 'hex'
    FLOAT = 'float'


@dc.dataclass()
class YamlNumberValue:
    type: YamlNumberType
    value: ta.Any
    text: str


def yaml_to_number(value: str) -> ta.Optional[YamlNumberValue]:
    num = _yaml_to_number(value)
    if isinstance(num, YamlError):
        return None

    return num


def _yaml_is_number(value: str) -> bool:
    num = _yaml_to_number(value)
    if isinstance(num, YamlError):
        # var numErr *strconv.NumError
        # if errors.As(err, &numErr) && errors.Is(numErr.Err, strconv.ErrRange) {
        #     return true

        return False

    return num is not None


def _yaml_to_number(value: str) -> YamlErrorOr[ta.Optional[YamlNumberValue]]:
    if not value:
        return None

    if value.startswith('_'):
        return None

    dot_count = value.count('.')
    if dot_count > 1:
        return None

    is_negative = value.startswith('-')
    normalized = value.lstrip('+').lstrip('-').replace('_', '')

    typ: YamlNumberType
    base = 0

    if normalized.startswith('0x'):
        normalized = normalized.lstrip('0x')
        base = 16
        typ = YamlNumberType.HEX
    elif normalized.startswith('0o'):
        normalized = normalized.lstrip('0o')
        base = 8
        typ = YamlNumberType.OCTET
    elif normalized.startswith('0b'):
        normalized = normalized.lstrip('0b')
        base = 2
        typ = YamlNumberType.BINARY
    elif normalized.startswith('0') and len(normalized) > 1 and dot_count == 0:
        base = 8
        typ = YamlNumberType.OCTET
    elif dot_count == 1:
        typ = YamlNumberType.FLOAT
    else:
        typ = YamlNumberType.DECIMAL
        base = 10

    text = normalized
    if is_negative:
        text = '-' + text

    v: ta.Any
    if typ == YamlNumberType.FLOAT:
        try:
            v = float(text)
        except ValueError as e:
            return yaml_error(e)
    else:
        try:
            v = int(text, base)
        except ValueError as e:
            return yaml_error(e)

    return YamlNumberValue(
        type=typ,
        value=v,
        text=text,
    )


##


# This is a subset of the formats permitted by the regular expression defined at http:#yaml.org/type/timestamp.html.
# Note that time.Parse cannot handle: "2001-12-14 21:59:43.10 -5" from the examples.
YAML_TIMESTAMP_FORMATS = (
    '%Y-%m-%dT%H:%M:%S.%fZ',  # RFC3339Nano
    '%Y-%m-%dt%H:%M:%S.%fZ',  # RFC3339Nano with lower-case "t"
    '%Y-%m-%d %H:%M:%S',      # DateTime
    '%Y-%m-%d',               # DateOnly

    # Not in examples, but to preserve backward compatibility by quoting time values
    '%H:%M',
)


def _yaml_is_timestamp(value: str) -> bool:
    for format_str in YAML_TIMESTAMP_FORMATS:
        try:
            datetime.datetime.strptime(value, format_str)  # noqa
            return True
        except ValueError:
            continue
    return False


##


# is_need_quoted checks whether the value needs quote for passed string or not
def _yaml_is_need_quoted(value: str) -> bool:
    if not value:
        return True

    if value in YamlKeywords.RESERVED_ENC_KEYWORD_MAP:
        return True

    if _yaml_is_number(value):
        return True

    if value == '-':
        return True

    if value[0] in ('*', '&', '[', '{', '}', ']', ',', '!', '|', '>', '%', '\'', '"', '@', ' ', '`'):
        return True

    if value[-1] in (':', ' '):
        return True

    if _yaml_is_timestamp(value):
        return True

    for i, c in enumerate(value):
        if c in ('#', '\\'):
            return True
        elif c in (':', '-'):
            if i + 1 < len(value) and value[i + 1] == ' ':
                return True

    return False


# literal_block_header detect literal block scalar header
def yaml_literal_block_header(value: str) -> str:
    lbc = yaml_detect_line_break_char(value)

    if lbc not in value:
        return ''
    elif value.endswith(lbc + lbc):
        return '|+'
    elif value.endswith(lbc):
        return '|'
    else:
        return '|-'


##


# new create reserved keyword token or number token and other string token.
def yaml_new_token(value: str, org: str, pos: 'YamlPosition') -> 'YamlToken':
    fn = YamlKeywords.RESERVED_KEYWORD_MAP.get(value)
    if fn is not None:
        return fn(value, org, pos)

    if (num := yaml_to_number(value)) is not None:
        tk = YamlToken(
            type=YamlTokenType.INTEGER,
            char_type=YamlCharType.MISCELLANEOUS,
            indicator=YamlIndicator.NOT,
            value=value,
            origin=org,
            position=pos,
        )
        if num.type == YamlNumberType.FLOAT:
            tk.type = YamlTokenType.FLOAT
        elif num.type == YamlNumberType.BINARY:
            tk.type = YamlTokenType.BINARY_INTEGER
        elif num.type == YamlNumberType.OCTET:
            tk.type = YamlTokenType.OCTET_INTEGER
        elif num.type == YamlNumberType.HEX:
            tk.type = YamlTokenType.HEX_INTEGER
        return tk

    return YamlTokenMakers.new_string(value, org, pos)


# Position type for position in YAML document
@dc.dataclass()
class YamlPosition:
    line: int
    column: int
    offset: int
    indent_num: int
    indent_level: int

    # String position to text
    def __str__(self) -> str:
        return f'[level:{self.indent_level:d},line:{self.line:d},column:{self.column:d},offset:{self.offset:d}]'


# Token type for token
@dc.dataclass()
@ta.final
class YamlToken:
    # Type is a token type.
    type: YamlTokenType
    # CharType is a character type.
    char_type: YamlCharType
    # Indicator is an indicator type.
    indicator: YamlIndicator
    # Value is a string extracted with only meaningful characters, with spaces and such removed.
    value: str
    # Origin is a string that stores the original text as-is.
    origin: str
    # Error keeps error message for InvalidToken.
    error: ta.Optional[YamlError] = None
    # Position is a token position.
    position: YamlPosition = dc.field(default_factory=dataclass_field_required('position'))
    # Next is a next token reference.
    next: ta.Optional['YamlToken'] = dc.field(default=None, repr=False)
    # Prev is a previous token reference.
    prev: ta.Optional['YamlToken'] = dc.field(default=None, repr=False)

    # previous_type previous token type
    def previous_type(self) -> YamlTokenType:
        if self.prev is not None:
            return self.prev.type

        return YamlTokenType.UNKNOWN

    # next_type next token type
    def next_type(self) -> YamlTokenType:
        if self.next is not None:
            return self.next.type

        return YamlTokenType.UNKNOWN

    # add_column append column number to current position of column
    @classmethod
    def add_column(cls, t: ta.Optional['YamlToken'], col: int) -> None:
        if t is None:
            return

        t.position.column += col

    # clone copy token ( preserve Prev/Next reference )
    @classmethod
    def clone(cls, t: ta.Optional['YamlToken']) -> ta.Optional['YamlToken']:
        if t is None:
            return None

        copied = copy.copy(t)
        if t.position is not None:
            pos = copy.copy(t.position)
            copied.position = pos

        return copied


##


# Tokens type of token collection
class YamlTokens(ta.List[YamlToken]):
    def invalid_token(self) -> ta.Optional[YamlToken]:
        for tt in self:
            if tt.type == YamlTokenType.INVALID:
                return tt
        return None

    def _add(self, tk: YamlToken) -> None:
        if not self:
            self.append(tk)
        else:
            last = self[-1]
            last.next = tk
            tk.prev = last
            self.append(tk)

    # add append new some tokens
    def add(self, *tks: YamlToken) -> None:
        for tk in tks:
            self._add(tk)


##


class YamlTokenMakers:  # noqa
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    # new_string create token for String
    @staticmethod
    def new_string(value: str, org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.STRING,
            char_type=YamlCharType.MISCELLANEOUS,
            indicator=YamlIndicator.NOT,
            value=value,
            origin=org,
            position=pos,
        )

    # new_sequence_entry create token for SequenceEntry
    @staticmethod
    def new_sequence_entry(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.SEQUENCE_ENTRY,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.BLOCK_STRUCTURE,
            value=YamlChars.SEQUENCE_ENTRY,
            origin=org,
            position=pos,
        )

    # new_mapping_key create token for MappingKey
    @staticmethod
    def new_mapping_key(pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.MAPPING_KEY,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.BLOCK_STRUCTURE,
            value=YamlChars.MAPPING_KEY,
            origin=YamlChars.MAPPING_KEY,
            position=pos,
        )

    # new_mapping_value create token for MappingValue
    @staticmethod
    def new_mapping_value(pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.MAPPING_VALUE,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.BLOCK_STRUCTURE,
            value=YamlChars.MAPPING_VALUE,
            origin=YamlChars.MAPPING_VALUE,
            position=pos,
        )

    # new_collect_entry create token for CollectEntry
    @staticmethod
    def new_collect_entry(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.COLLECT_ENTRY,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.FLOW_COLLECTION,
            value=YamlChars.COLLECT_ENTRY,
            origin=org,
            position=pos,
        )

    # new_sequence_start create token for SequenceStart
    @staticmethod
    def new_sequence_start(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.SEQUENCE_START,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.FLOW_COLLECTION,
            value=YamlChars.SEQUENCE_START,
            origin=org,
            position=pos,
        )

    # new_sequence_end create token for SequenceEnd
    @staticmethod
    def new_sequence_end(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.SEQUENCE_END,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.FLOW_COLLECTION,
            value=YamlChars.SEQUENCE_END,
            origin=org,
            position=pos,
        )

    # new_mapping_start create token for MappingStart
    @staticmethod
    def new_mapping_start(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.MAPPING_START,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.FLOW_COLLECTION,
            value=YamlChars.MAPPING_START,
            origin=org,
            position=pos,
        )

    # new_mapping_end create token for MappingEnd
    @staticmethod
    def new_mapping_end(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.MAPPING_END,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.FLOW_COLLECTION,
            value=YamlChars.MAPPING_END,
            origin=org,
            position=pos,
        )

    # new_comment create token for Comment
    @staticmethod
    def new_comment(value: str, org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.COMMENT,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.COMMENT,
            value=value,
            origin=org,
            position=pos,
        )

    # new_anchor create token for Anchor
    @staticmethod
    def new_anchor(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.ANCHOR,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.NODE_PROPERTY,
            value=YamlChars.ANCHOR,
            origin=org,
            position=pos,
        )

    # new_alias create token for Alias
    @staticmethod
    def new_alias(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.ALIAS,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.NODE_PROPERTY,
            value=YamlChars.ALIAS,
            origin=org,
            position=pos,
        )

    # new_tag create token for Tag
    @staticmethod
    def new_tag(value: str, org: str, pos: YamlPosition) -> YamlToken:
        fn = YAML_RESERVED_TAG_KEYWORD_MAP.get(value)
        if fn is not None:
            return fn(value, org, pos)

        return YamlToken(
            type=YamlTokenType.TAG,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.NODE_PROPERTY,
            value=value,
            origin=org,
            position=pos,
        )

    # new_literal create token for Literal
    @staticmethod
    def new_literal(value: str, org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.LITERAL,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.BLOCK_SCALAR,
            value=value,
            origin=org,
            position=pos,
        )

    # new_folded create token for Folded
    @staticmethod
    def new_folded(value: str, org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.FOLDED,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.BLOCK_SCALAR,
            value=value,
            origin=org,
            position=pos,
        )

    # new_single_quote create token for SingleQuote
    @staticmethod
    def new_single_quote(value: str, org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.SINGLE_QUOTE,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.QUOTED_SCALAR,
            value=value,
            origin=org,
            position=pos,
        )

    # new_double_quote create token for DoubleQuote
    @staticmethod
    def new_double_quote(value: str, org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.DOUBLE_QUOTE,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.QUOTED_SCALAR,
            value=value,
            origin=org,
            position=pos,
        )

    # new_directive create token for Directive
    @staticmethod
    def new_directive(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.DIRECTIVE,
            char_type=YamlCharType.INDICATOR,
            indicator=YamlIndicator.DIRECTIVE,
            value=YamlChars.DIRECTIVE,
            origin=org,
            position=pos,
        )

    # new_space create token for Space
    @staticmethod
    def new_space(pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.SPACE,
            char_type=YamlCharType.WHITESPACE,
            indicator=YamlIndicator.NOT,
            value=YamlChars.SPACE,
            origin=YamlChars.SPACE,
            position=pos,
        )

    # new_merge_key create token for MergeKey
    @staticmethod
    def new_merge_key(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.MERGE_KEY,
            char_type=YamlCharType.MISCELLANEOUS,
            indicator=YamlIndicator.NOT,
            value='<<',
            origin=org,
            position=pos,
        )

    # new_document_header create token for DocumentHeader
    @staticmethod
    def new_document_header(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.DOCUMENT_HEADER,
            char_type=YamlCharType.MISCELLANEOUS,
            indicator=YamlIndicator.NOT,
            value='---',
            origin=org,
            position=pos,
        )

    # new_document_end create token for DocumentEnd
    @staticmethod
    def new_document_end(org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.DOCUMENT_END,
            char_type=YamlCharType.MISCELLANEOUS,
            indicator=YamlIndicator.NOT,
            value='...',
            origin=org,
            position=pos,
        )

    @staticmethod
    def new_invalid(err: YamlError, org: str, pos: YamlPosition) -> YamlToken:
        return YamlToken(
            type=YamlTokenType.INVALID,
            char_type=YamlCharType.INVALID,
            indicator=YamlIndicator.NOT,
            value=org,
            origin=org,
            error=err,
            position=pos,
        )


##


# detect_line_break_char detect line break character in only one inside scalar content scope.
def yaml_detect_line_break_char(src: str) -> str:
    nc = src.count('\n')
    rc = src.count('\r')
    rnc = src.count('\r\n')
    if nc == rnc and rc == rnc:
        return '\r\n'
    elif rc > nc:
        return '\r'
    else:
        return '\n'



import typing as ta

from ... import check
from ... import collections as col
from ... import dispatch
from ... import lang
from .base import Op
from .matches import Match
from .ops import RuleRef


T = ta.TypeVar('T')


##


class OpMatchVisitor(lang.Abstract, ta.Generic[T]):
    @dispatch.method(
        instance_cache=True,
        strong_dispatch_cache=True,
    )
    def visit_op(self, o: Op, m: Match) -> T:
        raise TypeError(o)

    #

    def visit_match(self, m: Match) -> T:
        return self.visit_op(m.op, m)


##


class RuleMatchVisitor(lang.Abstract, ta.Generic[T]):
    _registry = col.AttrRegistry[ta.Callable, str]()

    @classmethod
    def register(cls, name):
        return cls._registry.register(name)

    @lang.cached_function
    @classmethod
    def _registry_cache(cls) -> col.AttrRegistryCache[ta.Callable, str, ta.Mapping[str, str]]:
        def prepare(_, dct):
            return col.make_map(((n, a) for a, (_, n) in dct.items()), strict=True)

        return col.WeakAttrRegistryCache(cls._registry, prepare)

    def visit_rule(self, name: str, m: Match) -> T:
        att = self._registry_cache().get(self.__class__)[name]
        return getattr(self, att)(m)

    #

    def visit_match(self, m: Match) -> T:
        return self.visit_rule(check.isinstance(m.op, RuleRef).name, m)



# ruff: noqa: UP007 UP043 UP045
# @omlish-lite
"""
TODO:
 - redirect
  - referrer header?
  - non-forwarded headers, host check, etc lol
 - 'check' kw becomes StatusCheckingMiddleware?
"""
import dataclasses as dc
import typing as ta
import urllib.parse

from ...lite.abstract import Abstract
from ...lite.check import check
from ..urls import parsed_url_replace
from .asyncs import AsyncHttpClient
from .asyncs import AsyncStreamHttpResponse
from .base import BaseHttpClient
from .base import BaseHttpResponse
from .base import HttpClientContext
from .base import HttpClientError
from .base import HttpRequest
from .sync import HttpClient
from .sync import StreamHttpResponse
from .sync import close_http_client_response


BaseHttpClientT = ta.TypeVar('BaseHttpClientT', bound=BaseHttpClient)


##


class HttpClientMiddleware(Abstract):
    def process_request(
            self,
            ctx: HttpClientContext,
            req: HttpRequest,
    ) -> HttpRequest:
        return req

    def process_response(
            self,
            ctx: HttpClientContext,
            req: HttpRequest,
            resp: BaseHttpResponse,
    ) -> ta.Union[BaseHttpResponse, HttpRequest]:
        return resp


class AbstractMiddlewareHttpClient(Abstract, ta.Generic[BaseHttpClientT]):
    def __init__(
            self,
            client: BaseHttpClientT,
            middlewares: ta.Iterable[HttpClientMiddleware],
    ) -> None:
        super().__init__()

        self._client = client
        self._middlewares = list(middlewares)

    def _process_request(
            self,
            ctx: HttpClientContext,
            req: HttpRequest,
    ) -> HttpRequest:
        for mw in self._middlewares:
            req = mw.process_request(ctx, req)
        return req

    def _process_response(
            self,
            ctx: HttpClientContext,
            req: HttpRequest,
            resp: BaseHttpResponse,
    ) -> ta.Union[BaseHttpResponse, HttpRequest]:
        for mw in self._middlewares:
            nxt = mw.process_response(ctx, req, resp)
            if isinstance(nxt, HttpRequest):
                return nxt
            else:
                resp = nxt
        return resp


#


class MiddlewareHttpClient(AbstractMiddlewareHttpClient[HttpClient], HttpClient):
    def _stream_request(self, ctx: HttpClientContext, req: HttpRequest) -> StreamHttpResponse:
        while True:
            req = self._process_request(ctx, req)

            resp = self._client.stream_request(req, context=ctx)

            try:
                out = self._process_response(ctx, req, resp)

                if isinstance(out, HttpRequest):
                    close_http_client_response(resp)
                    req = out
                    continue

                elif isinstance(out, StreamHttpResponse):
                    return out

                else:
                    raise TypeError(out)  # noqa

            except Exception:
                close_http_client_response(resp)
                raise

        raise RuntimeError


class MiddlewareAsyncHttpClient(AbstractMiddlewareHttpClient[AsyncHttpClient], AsyncHttpClient):
    def _stream_request(self, ctx: HttpClientContext, req: HttpRequest) -> ta.Awaitable[AsyncStreamHttpResponse]:
        return self._client.stream_request(self._process_request(ctx, req))


##


class TooManyRedirectsHttpClientError(HttpClientError):
    pass


class RedirectHandlingHttpClientMiddleware(HttpClientMiddleware):
    DEFAULT_MAX_REDIRECTS: ta.ClassVar[int] = 5

    def __init__(
            self,
            *,
            max_redirects: ta.Optional[int] = None,
    ) -> None:
        super().__init__()

        if max_redirects is None:
            max_redirects = self.DEFAULT_MAX_REDIRECTS
        self._max_redirects = max_redirects

    @dc.dataclass()
    class _State:
        num_redirects: int = 0

    def _get_state(self, ctx: HttpClientContext) -> _State:
        try:
            return ctx._dct[self._State]  # noqa
        except KeyError:
            ret = ctx._dct[self._State] = self._State()  # noqa
            return ret

    def process_response(
            self,
            ctx: HttpClientContext,
            req: HttpRequest,
            resp: BaseHttpResponse,
    ) -> ta.Union[BaseHttpResponse, HttpRequest]:  # noqa
        if resp.status == 302:
            st = self._get_state(ctx)
            if st.num_redirects >= self._max_redirects:
                raise TooManyRedirectsHttpClientError
            st.num_redirects += 1

            rd_url = check.not_none(resp.headers).single['location']

            rd_purl = urllib.parse.urlparse(rd_url)
            if not rd_purl.netloc:
                rq_purl = urllib.parse.urlparse(req.url)
                rd_purl = parsed_url_replace(
                    rd_purl,
                    scheme=rq_purl.scheme,
                    netloc=rq_purl.netloc,
                )
                rd_url = urllib.parse.urlunparse(rd_purl)

            return dc.replace(req, url=rd_url)

        return resp



"""
TODO:
 - origin tracking?
 - minor opts:
  - merge concat(range, range)
"""
import abc
import re
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from .base import CompositeOp
from .base import Op
from .grammars import Channel
from .grammars import Grammar
from .grammars import Rule
from .internal import Regex
from .ops import CaseInsensitiveStringLiteral
from .ops import Concat
from .ops import Either
from .ops import RangeLiteral
from .ops import Repeat
from .ops import RuleRef
from .ops import StringLiteral


##


@dc.dataclass(frozen=True)
class _RegexItem(lang.Abstract):
    @property
    @abc.abstractmethod
    def pat(self) -> str:
        raise NotImplementedError

    @classmethod
    def of_op(cls, op: Op) -> ta.Optional['_RegexItem']:
        if isinstance(op, StringLiteral):
            return _StringLiteralRegexItem(op.value)

        elif isinstance(op, CaseInsensitiveStringLiteral):
            return _CaseInsensitiveStringLiteralRegexItem(op.value)

        elif isinstance(op, RangeLiteral):
            lo = re.escape(op.value.lo)
            hi = re.escape(op.value.hi)
            return _RegexRegexItem(f'[{lo}-{hi}]')

        elif isinstance(op, Regex):
            return _RegexRegexItem(op.pat.pattern)

        else:
            return None

    @classmethod
    def of(cls, obj: ta.Union['_RegexItem', Op, None]) -> ta.Optional['_RegexItem']:
        if obj is None:
            return None

        elif isinstance(obj, _RegexItem):
            return obj

        elif isinstance(obj, Op):
            return cls.of_op(obj)

        else:
            raise TypeError(obj)


@dc.dataclass(frozen=True)
class _StringLiteralRegexItem(_RegexItem, lang.Final):
    s: str

    @property
    def pat(self) -> str:
        return re.escape(self.s)


@dc.dataclass(frozen=True)
class _CaseInsensitiveStringLiteralRegexItem(_RegexItem, lang.Final):
    s: str

    @property
    def pat(self) -> str:
        return f'(?i:{re.escape(self.s)})'


@dc.dataclass(frozen=True)
class _RegexRegexItem(_RegexItem, lang.Final):
    ps: str

    @property
    def pat(self) -> str:
        return self.ps


def _regex_item_transform_op(op: Op) -> _RegexItem | None:
    if isinstance(op, (StringLiteral, CaseInsensitiveStringLiteral, Regex)):
        return None

    elif isinstance(op, RangeLiteral):
        # Unlike other leafs we eagerly transform RangeLiteral to a regex as it's probably faster than the python impl,
        # even alone.
        return _RegexItem.of_op(op)

    elif isinstance(op, RuleRef):
        return None

    elif isinstance(op, Concat):
        children = [_regex_item_transform_op(child) or _RegexItem.of(child) for child in op.children]
        if all(ca is not None for ca in children):
            return _RegexRegexItem(''.join(check.not_none(ca).pat for ca in children))

        if not any(ca is not None for ca in children):
            return None

        # FIXME: merge adjacent
        return None

    elif isinstance(op, Repeat):
        child = _RegexItem.of(_regex_item_transform_op(op.child))
        if child is None:
            return None

        # Wrap the child pattern in a non-capturing group if needed to ensure correct quantification. A pattern needs
        # wrapping if it contains multiple elements or operators (e.g., 'ab', 'a|b'). Single character classes [a-z] and
        # single escaped chars don't need wrapping.
        if (
                len(child_pat := child.pat) > 1 and
                not (child_pat.startswith('[') and child_pat.endswith(']'))
        ):
            child_pat = f'(?:{child_pat})'

        times = op.times
        if times.min == 0 and times.max is None:
            quantifier = '*'
        elif times.min == 1 and times.max is None:
            quantifier = '+'
        elif times.min == 0 and times.max == 1:
            quantifier = '?'
        elif times.max is None:
            quantifier = f'{{{times.min},}}'
        elif times.min == times.max:
            quantifier = f'{{{times.min}}}'
        else:
            quantifier = f'{{{times.min},{times.max}}}'

        return _RegexRegexItem(child_pat + quantifier)

    elif isinstance(op, Either):
        # Only convert Either if first_match is True, as regex alternation uses first-match semantics. ABNF Either with
        # first_match=False uses longest-match semantics, which differs from regex.
        if not op.first_match:
            return None

        children = [_regex_item_transform_op(child) or _RegexItem.of(child) for child in op.children]
        if all(ca is not None for ca in children):
            # Build regex alternation. Use a capturing group for the alternation
            return _RegexRegexItem(''.join([
                '(',
                '|'.join(check.not_none(ca).pat for ca in children),
                ')',
            ]))

        if not any(ca is not None for ca in children):
            return None

        # FIXME: merge adjacent
        return None

    else:
        raise TypeError(op)


def _regex_transform_op(op: Op) -> Op:
    v = _regex_item_transform_op(op)

    if v is None:
        return op

    elif isinstance(v, _RegexItem):
        return Regex(re.compile(v.pat))

    else:
        raise TypeError(v)


##


def optimize_op(op: Op) -> Op:
    op = _regex_transform_op(op)

    return op


##


def _inline_rules(fn: ta.Callable[[Rule], bool], gram: Grammar) -> Grammar:
    cur_rule: Rule
    inlined_rules: dict[str, Op] = {}

    def rec_op(op: Op) -> Op:
        if isinstance(op, RuleRef):
            if op.name_f == cur_rule.name_f:
                return op

            if (r := gram.rule(op.name)) is None or not fn(r):
                return op

            try:
                return inlined_rules[r.name]
            except KeyError:
                pass

            inlined_rules[op.name] = op
            i_op = rec_op(r.op)
            inlined_rules[op.name] = i_op

            return op.coalesce(i_op)

        elif isinstance(op, CompositeOp):
            return op.replace_children(*map(rec_op, op.children))

        else:
            return op

    new_rules: list[Rule] = []
    for rule in gram.rules:
        cur_rule = rule
        new_rules.append(rule.replace_op(rec_op(rule.op)))

    return gram.replace_rules(*new_rules)


##


def optimize_grammar(
        gram: Grammar,
        *,
        inline_channels: ta.Container[Channel] | None = (Channel.SPACE,),
) -> Grammar:
    if inline_channels:
        gram = _inline_rules(lambda r: r.channel in inline_channels, gram)

    gram = gram.replace_rules(*[
        r.replace_op(optimize_op(r.op))
        for r in gram.rules
    ])

    return gram



# ruff: noqa: I001
from .rlock import (  # noqa
    AsyncioRLock as RLock,
)

from .threads import (  # noqa
    AsyncioToThread as ToThread,
)

from .utils import (  # noqa
    asyncio_ensure_task as ensure_task,

    asyncio_once as once,

    drain_asyncio_tasks as drain_tasks,
    draining_asyncio_tasks as draining_tasks,

    asyncio_wait_concurrent as wait_concurrent,
    asyncio_wait_maybe_concurrent as wait_maybe_concurrent,
)



import re
import typing as ta

from ... import lang
from .ops import LeafOp
from .ops import Op


##


class InternalOp(Op, lang.Abstract):
    pass


##


@ta.final
class Regex(InternalOp, LeafOp, lang.Final):
    def __init__(self, pat: re.Pattern) -> None:
        super().__init__()

        self._pat = pat

    @property
    def pat(self) -> re.Pattern:
        return self._pat

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._pat.pattern!r})'



"""
TODO:
 - end-of-line   = ( cr lf / cr / lf )

See:
 - https://github.com/florimondmanca/httpx-sse/blob/master/src/httpx_sse/_decoders.py
"""
import string
import typing as ta

from .. import dataclasses as dc
from .. import lang


##


class SseDecoderOutput(lang.Abstract):
    pass


@dc.dataclass(frozen=True)
class SseComment(SseDecoderOutput, lang.Final):
    data: bytes


SseEventId: ta.TypeAlias = bytes


@dc.dataclass(frozen=True)
class SseEvent(SseDecoderOutput, lang.Final):
    type: bytes
    data: bytes
    last_id: SseEventId = dc.xfield(b'', repr_fn=lang.truthy_repr)


_DIGIT_BYTES = string.digits.encode('ascii')


class SseDecoder:
    """https://html.spec.whatwg.org/multipage/server-sent-events.html#event-stream-interpretation"""

    def __init__(self) -> None:
        super().__init__()

        self._reset()
        self._last_event_id = b''
        self._reconnection_time: int | None = None

    _event_type: bytes
    _data: list[bytes]

    def _reset(self) -> None:
        self._event_type = b'message'
        self._data = []

    def _process_field(self, name: bytes, value: bytes) -> None:
        if name == b'event':
            self._event_type = value

        elif name == b'data':
            self._data.append(value)

        elif name == b'id':
            if 0 not in value:
                self._last_event_id = value

        elif name == b'retry':
            if all(c in _DIGIT_BYTES for c in value):
                self._reconnection_time = int(value)

    def _dispatch_event(self) -> SseEvent:
        data = b''.join(lang.interleave(self._data, b'\n'))

        e = SseEvent(
            type=self._event_type,
            data=data,
            last_id=self._last_event_id,
        )

        self._reset()

        return e

    def process_line(self, line: bytes) -> ta.Iterator[SseDecoderOutput]:
        if b'\r' in line or b'\n' in line:
            raise ValueError(line)

        if not line:
            yield self._dispatch_event()

        elif line[0] == b':'[0]:
            yield SseComment(line)

        elif (c := line.find(b':')) >= 0:
            d = c + 1
            if len(line) > d and line[d] == b' '[0]:
                d += 1
            self._process_field(line[:c], line[d:])

        else:
            self._process_field(line, b'')



# ruff: noqa: UP006 UP043 UP045
# @omlish-lite
import contextlib
import threading
import typing as ta


T = ta.TypeVar('T')


##


class ObjectPool(ta.Generic[T]):
    def __init__(self, new: ta.Optional[ta.Callable[[], T]] = None) -> None:
        super().__init__()

        self._new = new

        self._lock = threading.Lock()
        self._pool: ta.List[T] = []
        self._closed: bool = False

    #

    class ClosedError(Exception):
        pass

    def _check_not_closed(self) -> None:
        if self._closed:
            raise ObjectPool.ClosedError

    def close(self) -> None:
        with self._lock:
            self._closed = True

    #

    class EmptyError(Exception):
        pass

    def get(self, new: ta.Optional[ta.Callable[[], T]] = None) -> T:
        with self._lock:
            self._check_not_closed()
            if self._pool:
                return self._pool.pop()

        if new is None:
            new = self._new
        if new is not None:
            return new()

        raise ObjectPool.EmptyError

    def put(self, obj: T) -> None:
        with self._lock:
            self._check_not_closed()
            self._pool.append(obj)

    def acquire(self, new: ta.Optional[ta.Callable[[], T]] = None) -> ta.ContextManager[T]:
        @contextlib.contextmanager
        def inner():
            o = self.get(new)
            try:
                yield o
            finally:
                self.put(o)

        return inner()

    def drain(self) -> ta.List[T]:
        with self._lock:
            out = self._pool
            self._pool = []

        return out

    #

    def manage(
            self,
            drain: ta.Optional[ta.Callable[[T], None]] = None,
    ) -> ta.ContextManager['ObjectPool[T]']:
        @contextlib.contextmanager
        def inner():
            with self._lock:
                self._check_not_closed()

            try:
                yield self

            finally:
                self.close()

                if drain is not None:
                    for o in self.drain():
                        drain(o)

        return inner()



# ruff: noqa: I001
from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from . import api  # noqa

    from .api import (  # noqa
        Querier,
        Rows,
        Transaction,
        Conn,
        Db,
        Adapter,

        Column,
        Columns,

        exec,  # noqa

        query,
        query_all,
        query_first,
        query_opt_first,
        query_one,
        query_opt_one,
        query_scalar,
        query_maybe_scalar,

        QueryMode,
        Query,

        Row,
    )

    #

    from . import dbapi  # noqa

    #

    from . import queries  # noqa

    from .queries import (  # noqa
        Q,
    )

    #

    from . import tabledefs  # noqa

    from . import tabledefs as td  # noqa

    #

    from .dbs import (  # noqa
        DbType,
        DbTypes,

        DbLoc,
        UrlDbLoc,
        HostDbLoc,

        DbSpec,
    )

    from .params import (  # noqa
        ParamKey,

        SequencePreparedParams,
        MappingPreparedParams,

        ParamsPreparer,
        SequenceParamsPreparer,
        MappingParamsPreparer,

        LinearParamsPreparer,
        NumericParamsPreparer,
        NamedParamsPreparer,

        ParamStyle,
        make_params_preparer,

        UnconsumedParamsError,
        check_params_consumed,

        substitute_params,
    )

    from .qualifiedname import (  # noqa
        QualifiedName,
        qn,
    )



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import functools
import typing as ta


##


class UnknownHttpVersionError(Exception):
    pass


@ta.final
@functools.total_ordering
class HttpVersion:
    def __init__(self, major: int, minor: int) -> None:
        self._major = major
        self._minor = minor

        self._parts = parts = (major, minor)

        self._hash = hash(parts)

        self._str = f'HTTP/{major}.{minor}'
        self._short_str = f'{major}.{minor}'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.major}, {self.minor})'

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other: object) -> ta.Any:
        if not isinstance(other, HttpVersion):
            return NotImplemented
        return self._parts == other._parts

    def __lt__(self, other: object) -> ta.Any:
        if not isinstance(other, HttpVersion):
            return NotImplemented
        return self._parts < other._parts

    @property
    def major(self) -> int:
        return self._major

    @property
    def minor(self) -> int:
        return self._minor

    def __str__(self) -> str:
        return self._str

    @property
    def short_str(self) -> str:
        return self._short_str

    def __iter__(self) -> ta.Iterator[int]:
        return iter(self._parts)


@ta.final
class HttpVersions:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    HTTP_0_9 = HttpVersion(0, 9)
    HTTP_1_0 = HttpVersion(1, 0)
    HTTP_1_1 = HttpVersion(1, 1)
    HTTP_2_0 = HttpVersion(2, 0)

    _FROM_STR: ta.ClassVar[ta.Mapping[str, HttpVersion]] = {
        str(v): v for v in [
            HTTP_0_9,
            HTTP_1_0,
            HTTP_1_1,
            HTTP_2_0,
        ]
    }

    @classmethod
    def from_str(cls, s: str) -> HttpVersion:
        try:
            return cls._FROM_STR[s]
        except KeyError:
            raise UnknownHttpVersionError(s) from None

    @classmethod
    def of(cls, o: ta.Union[HttpVersion, str]) -> HttpVersion:
        if isinstance(o, HttpVersion):
            return o
        elif isinstance(o, str):
            return cls.from_str(o)
        else:
            raise TypeError(o)



import dataclasses as dc
import typing as ta

from ..funcs import pairs as fps
from .base import EagerCodec


I = ta.TypeVar('I')
O = ta.TypeVar('O')


##


@dc.dataclass(frozen=True)
class FnPairEagerCodec(EagerCodec[I, O]):
    fp: fps.FnPair[I, O]

    def encode(self, i: I) -> O:
        return self.fp.forward(i)

    def decode(self, o: O) -> I:
        return self.fp.backward(o)

    @classmethod
    def of(
            cls,
            encode: ta.Callable[[I], O],
            decode: ta.Callable[[O], I],
    ) -> 'FnPairEagerCodec[I, O]':
        return cls(fps.of(encode, decode))


def of_pair(fp: fps.FnPair[I, O]) -> FnPairEagerCodec[I, O]:
    return FnPairEagerCodec(fp)


def of(
        encode: ta.Callable[[I], O],
        decode: ta.Callable[[O], I],
) -> FnPairEagerCodec[I, O]:
    return FnPairEagerCodec(fps.of(encode, decode))



"""
Usable as a jetbrains external tool:
  `om docwrap -i "$FilePath$" -s "$SelectionStartLine$" -e "$SelectionEndLine$"`

TODO:
 - fix 'stealing trailing line'
 - -> omdev.tools
 - at least file extension awareness, preserve say # prefixes in py comment blocks
  - maybe treesitter? or just special case py
"""
import argparse
import sys
import typing as ta

from .api import docwrap
from .rendering import render


##


def _main(argv: ta.Sequence[str] | None = None) -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('file', nargs='?')

    parser.add_argument('-w', '--width', type=int, default=120)

    parser.add_argument('-s', '--start-line', type=int)
    parser.add_argument('-e', '--end-line', type=int)
    parser.add_argument('-i', '--in-place', action='store_true')

    args = parser.parse_args(argv)

    #

    if args.file:
        with open(args.file) as f:
            in_txt = f.read()
    else:
        if args.in_place:
            raise ValueError('Cannot use --in-place without specifying a file')
        in_txt = sys.stdin.read()

    in_lines = in_txt.splitlines()

    #

    if args.start_line is not None and args.end_line is not None:
        if args.start_line > args.end_line:
            raise ValueError('Start line cannot be greater than end line')
    if args.start_line is not None:
        if args.start_line < 1:
            raise ValueError('Start line cannot be less than 1')
        start_line = args.start_line - 1
    else:
        start_line = 0
    if args.end_line is not None:
        if args.end_line < 1:
            raise ValueError('End line cannot be less than 1')
        end_line = args.end_line - 1
    else:
        end_line = len(in_lines) - 1

    #

    in_part = '\n'.join(in_lines[start_line:end_line + 1])

    root = docwrap(
        in_part,
        width=args.width,
    )

    out_part = render(root)

    out_txt = '\n'.join([
        *in_lines[:start_line],
        out_part,
        *in_lines[end_line + 1:],
        '',
    ])

    #

    if args.in_place:
        with open(args.file, 'w') as f:
            f.write(out_txt)
    else:
        print(out_txt)


if __name__ == '__main__':
    _main()



import contextvars
import dataclasses as dc

from ... import lang
from ...logs import all as logs
from .. import asgi
from .. import sessions
from .markers import AppMarker
from .markers import AppMarkerProcessor
from .markers import append_app_marker


log = logs.get_module_logger(globals())


##


SESSION: contextvars.ContextVar[sessions.Session] = contextvars.ContextVar('session')


class _WithSessionAppMarker(AppMarker, lang.Singleton, lang.Final):
    pass


def with_session(fn):
    return append_app_marker(fn, _WithSessionAppMarker())


@dc.dataclass(frozen=True)
class _WithSessionAppMarkerProcessor(AppMarkerProcessor):
    _ss: sessions.CookieSessionStore

    async def _wrap(self, fn: asgi.App, scope: asgi.Scope, recv: asgi.Recv, send: asgi.Send) -> None:
        async def _send(obj):
            if obj['type'] == 'http.response.start':
                out_session = SESSION.get()
                obj = {
                    **obj,
                    'headers': [
                        *obj.get('headers', []),
                        *self._ss.build_headers(out_session),
                    ],
                }

            await send(obj)

        in_session = self._ss.extract(scope)
        with lang.context_var_setting(SESSION, in_session):
            await fn(scope, recv, _send)

    def process_app(self, app: asgi.App) -> asgi.App:
        return lang.decorator(self._wrap)(app)  # noqa



import abc
import contextlib
import typing as ta

from ... import lang
from ..headers import CanHttpHeaders
from .asyncs import AsyncHttpClient
from .base import HttpClientContext
from .base import HttpRequest
from .base import HttpResponse
from .sync import HttpClient


with lang.auto_proxy_import(globals()):
    from . import httpx as _httpx
    from . import urllib as _urllib


C = ta.TypeVar('C')
R = ta.TypeVar('R')


##


class _DefaultRequester(lang.Abstract, ta.Generic[C, R]):
    def __call__(
            self,
            url: str,
            method: str | None = None,
            *,
            headers: CanHttpHeaders | None = None,
            data: bytes | str | None = None,

            timeout_s: float | None = None,

            context: HttpClientContext | None = None,
            check: bool = False,
            client: C | None = None,  # noqa

            **kwargs: ta.Any,
    ) -> R:
        request = HttpRequest(  # noqa
            url,
            method=method,

            headers=headers,
            data=data,

            timeout_s=timeout_s,

            **kwargs,
        )

        return self._do(
            request,
            context=context,
            check=check,
            client=client,
        )

    @abc.abstractmethod
    def _do(
            self,
            request: HttpRequest,  # noqa
            *,
            context: HttpClientContext | None = None,
            check: bool = False,
            client: C | None = None,  # noqa
    ) -> R:
        raise NotImplementedError


##


def _default_client() -> HttpClient:
    return _urllib.UrllibHttpClient()


def client() -> HttpClient:
    return _default_client()


@contextlib.contextmanager
def manage_client(client: HttpClient | None = None) -> ta.Generator[HttpClient]:  # noqa
    if client is not None:
        yield client

    else:
        with _default_client() as client:  # noqa
            yield client


#


class _BaseSyncDefaultRequester(_DefaultRequester[HttpClient, R], lang.Abstract, ta.Generic[R]):
    def _do(
            self,
            request: HttpRequest,  # noqa
            *,
            context: HttpClientContext | None = None,
            check: bool = False,
            client: HttpClient | None = None,  # noqa
    ) -> R:
        if context is None:
            context = HttpClientContext()

        if client is not None:
            return self._do_(
                client,
                context,
                request,
                check=check,
            )

        else:
            with _default_client() as client:  # noqa
                return self._do_(
                    client,
                    context,
                    request,
                    check=check,
                )

    @abc.abstractmethod
    def _do_(
            self,
            client: HttpClient,  # noqa
            context: HttpClientContext,
            request: HttpRequest,  # noqa
            *,
            check: bool = False,  # noqa
    ) -> R:
        raise NotImplementedError


class _SyncDefaultRequester(_BaseSyncDefaultRequester[HttpResponse]):
    def _do_(
            self,
            client: HttpClient,  # noqa
            context: HttpClientContext,
            request: HttpRequest,  # noqa
            *,
            check: bool = False,  # noqa
    ) -> HttpResponse:
        return client.request(
            request,
            context=context,
            check=check,
        )


request = _SyncDefaultRequester()


##


def _default_async_client() -> AsyncHttpClient:
    return _httpx.HttpxAsyncHttpClient()


def async_client() -> AsyncHttpClient:
    return _default_async_client()


@contextlib.asynccontextmanager
async def manage_async_client(client: AsyncHttpClient | None = None) -> ta.AsyncGenerator[AsyncHttpClient]:  # noqa
    if client is not None:
        yield client

    else:
        async with _default_async_client() as client:  # noqa
            yield client


#


class _BaseAsyncDefaultRequester(_DefaultRequester[AsyncHttpClient, ta.Awaitable[R]], lang.Abstract, ta.Generic[R]):
    async def _do(
            self,
            request: HttpRequest,  # noqa
            *,
            context: HttpClientContext | None = None,
            check: bool = False,
            client: AsyncHttpClient | None = None,  # noqa
    ) -> R:
        if context is None:
            context = HttpClientContext()

        if client is not None:
            return await self._do_(
                client,
                context,
                request,
                check=check,
            )

        else:
            async with _default_async_client() as client:  # noqa
                return await self._do_(
                    client,
                    context,
                    request,
                    check=check,
                )

    @abc.abstractmethod
    def _do_(
            self,
            client: AsyncHttpClient,  # noqa
            context: HttpClientContext,
            request: HttpRequest,  # noqa
            *,
            check: bool = False,  # noqa
    ) -> ta.Awaitable[R]:
        raise NotImplementedError


class _AsyncDefaultRequester(_BaseAsyncDefaultRequester[HttpResponse]):
    async def _do_(
            self,
            client: AsyncHttpClient,  # noqa
            context: HttpClientContext,
            request: HttpRequest,  # noqa
            *,
            check: bool = False,
    ) -> HttpResponse:  # noqa
        return await client.request(
            request,
            context=context,
            check=check,
        )


async_request = _AsyncDefaultRequester()



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import typing as ta

from .base import BaseByteStreamBufferLike
from .direct import _EMPTY_DIRECT_BYTE_STREAM_BUFFER_VIEW
from .direct import DirectByteStreamBufferView
from .errors import BufferTooLargeByteStreamBufferError
from .errors import NoOutstandingReserveByteStreamBufferError
from .errors import OutstandingReserveByteStreamBufferError
from .types import BytesLike
from .types import ByteStreamBufferView
from .types import MutableByteStreamBuffer
from .utils import ByteStreamBuffers


T = ta.TypeVar('T')


##


class SegmentedByteStreamBufferView(BaseByteStreamBufferLike, ByteStreamBufferView):
    """
    A read-only, possibly non-contiguous view over a sequence of byte segments.

    This is intended to be produced by `SegmentedByteStreamBuffer.split_to()` without copying.
    """

    def __init__(self, segs: ta.Sequence[memoryview]) -> None:
        super().__init__()

        self._segs = tuple(segs)
        for mv in self._segs:
            if (mvl := len(mv)) < 1:
                raise ValueError('Empty segment')
            self._len += mvl

    @classmethod
    def of_opt(cls, segs: ta.Sequence[memoryview]) -> ta.Optional['SegmentedByteStreamBufferView']:
        if not segs:
            return None
        return cls(segs)

    @classmethod
    def or_else(cls, segs: ta.Sequence[memoryview], default: T) -> ta.Union['SegmentedByteStreamBufferView', T]:
        if not segs:
            return default
        return cls(segs)

    _len = 0

    def __len__(self) -> int:
        return self._len

    def peek(self) -> memoryview:
        if not self._segs:
            return memoryview(b'')
        return self._segs[0]

    def segments(self) -> ta.Sequence[memoryview]:
        return self._segs

    def tobytes(self) -> bytes:
        if not self._segs:
            return b''
        if len(self._segs) == 1:
            return ByteStreamBuffers.memoryview_to_bytes(self._segs[0])
        return b''.join(ByteStreamBuffers.memoryview_to_bytes(mv) for mv in self._segs)


class SegmentedByteStreamBuffer(BaseByteStreamBufferLike, MutableByteStreamBuffer):
    """
    A segmented, consumption-oriented bytes buffer.

    Internally stores a list of `bytes`/`bytearray` segments plus a head offset. Exposes readable data as `memoryview`
    segments without copying.

    Optional "chunked writes":
      - If chunk_size > 0, small writes are accumulated into a lazily-allocated active bytearray "chunk" up to
        chunk_size.
      - Writes >= chunk_size are stored as their own segments (after flushing any active chunk).
      - On flush, the active chunk is kept as a bytearray segment iff it is at least `chunk_compact_threshold` full;
        otherwise it is materialized as bytes to avoid pinning a large capacity for tiny content.

    Reserve/commit:
      - If chunk_size > 0 and reserve(n) fits in the active chunk, the reservation is carved from the active chunk.
        Reserved bytes are not readable until commit().
      - If reserve(n) does not fit, the active chunk is flushed first.
      - If n <= chunk_size after flushing, the reservation is served from a new active chunk (so the remainder becomes
        the next active chunk).
      - If n > chunk_size, reserve allocates a dedicated buffer and on commit it is "closed" (it does not become the
        next active chunk).

    Important exported-view caveat:
      - reserve() returns a memoryview. As long as any exported memoryview exists, the underlying bytearray must not be
        resized, or Python will raise BufferError. Therefore the active chunk bytearray is *fixed capacity*
        (len==chunk_size) and we track "used" bytes separately, writing via slice assignment rather than extend().
    """

    def __init__(
            self,
            *,
            max_size: ta.Optional[int] = None,
            chunk_size: int = 0,
            chunk_compact_threshold: float = .25,
    ) -> None:
        super().__init__()

        self._segs: ta.List[ta.Union[bytes, bytearray]] = []

        self._max_size = None if max_size is None else int(max_size)

        if chunk_size < 0:
            raise ValueError(chunk_size)
        self._chunk_size = chunk_size

        if not (0.0 <= chunk_compact_threshold <= 1.0):
            raise ValueError(chunk_compact_threshold)
        self._chunk_compact_threshold = chunk_compact_threshold

        self._active: ta.Optional[bytearray] = None
        self._active_used = 0

    @property
    def max_size(self) -> ta.Optional[int]:
        return self._max_size

    _head_off = 0
    _len = 0

    _reserved: ta.Optional[bytearray] = None
    _reserved_len = 0
    _reserved_in_active = False

    #

    def __len__(self) -> int:
        return self._len

    def _active_readable_len(self) -> int:
        if self._active is None:
            return 0
        if self._reserved_in_active and self._reserved is not None:
            tail = self._reserved_len
        else:
            tail = 0
        rl = self._active_used - tail
        return rl if rl > 0 else 0

    def peek(self) -> memoryview:
        if not self._segs:
            return memoryview(b'')

        s0 = self._segs[0]
        mv = memoryview(s0)
        if self._head_off:
            mv = mv[self._head_off:]

        if s0 is self._active:
            # Active is only meaningful by _active_used, not len(bytearray).
            rl = self._active_readable_len()
            if self._head_off >= rl:
                return memoryview(b'')
            mv = memoryview(self._active)[self._head_off:rl]
            return mv

        return mv

    def segments(self) -> ta.Sequence[memoryview]:
        if not self._segs:
            return ()

        out: ta.List[memoryview] = []

        last_i = len(self._segs) - 1
        for i, s in enumerate(self._segs):
            if s is self._active and i == last_i:
                # Active chunk: create fresh view with readable length.
                rl = self._active_readable_len()
                if not i:
                    # Active is also first segment; apply head_off.
                    if self._head_off >= rl:
                        continue
                    mv = memoryview(self._active)[self._head_off:rl]
                else:
                    if rl <= 0:
                        continue
                    mv = memoryview(self._active)[:rl]
            else:
                # Non-active segment.
                mv = memoryview(s)
                if not i and self._head_off:
                    mv = mv[self._head_off:]

            if len(mv):
                out.append(mv)

        return out

    #

    def _ensure_active(self) -> bytearray:
        if self._chunk_size <= 0:
            raise RuntimeError('no active chunk without chunk_size')

        a = self._active
        if a is None:
            a = bytearray(self._chunk_size)  # fixed capacity
            self._segs.append(a)
            self._active = a
            self._active_used = 0

        return a

    def _flush_active(self) -> None:
        if (a := self._active) is None:
            return

        if self._reserved_in_active:
            raise OutstandingReserveByteStreamBufferError('outstanding reserve')

        if (used := self._active_used) <= 0:
            if self._segs and self._segs[-1] is a:
                self._segs.pop()
            self._active = None
            self._active_used = 0
            return

        # If under threshold, always bytes() to avoid pinning.
        if self._chunk_size and (float(used) / float(self._chunk_size)) < self._chunk_compact_threshold:
            if not self._segs or self._segs[-1] is not a:
                raise RuntimeError('active not at tail')
            self._segs[-1] = ByteStreamBuffers.memoryview_to_bytes(memoryview(a)[:used])

        else:
            # Try to shrink in-place to used bytes. If exported views exist, this can BufferError; fall back to bytes()
            # in that case.
            if not self._segs or self._segs[-1] is not a:
                raise RuntimeError('active not at tail')
            try:
                del a[used:]  # may raise BufferError if any exports exist
            except BufferError:
                self._segs[-1] = ByteStreamBuffers.memoryview_to_bytes(memoryview(a)[:used])

        self._active = None
        self._active_used = 0

    def write(self, data: BytesLike, /) -> None:
        if not data:
            return
        if isinstance(data, memoryview):
            data = ByteStreamBuffers.memoryview_to_bytes(data)  # noqa
        # elif isinstance(data, bytearray):
        #     pass
        # else:
        #     pass

        dl = len(data)

        if self._max_size is not None and self._len + dl > self._max_size:
            raise BufferTooLargeByteStreamBufferError('buffer exceeded max_size')

        if self._chunk_size <= 0:
            self._segs.append(data)
            self._len += dl
            return

        if self._reserved_in_active:
            raise OutstandingReserveByteStreamBufferError('outstanding reserve')

        if dl >= self._chunk_size:
            self._flush_active()
            self._segs.append(data)
            self._len += dl
            return

        a = self._ensure_active()
        if self._active_used + dl > self._chunk_size:
            self._flush_active()
            a = self._ensure_active()

        # Copy into fixed-capacity buffer; do not resize.
        memoryview(a)[self._active_used:self._active_used + dl] = data
        self._active_used += dl
        self._len += dl

    def reserve(self, n: int, /) -> memoryview:
        if n < 0:
            raise ValueError(n)
        if self._reserved is not None:
            raise OutstandingReserveByteStreamBufferError('outstanding reserve')

        if self._chunk_size <= 0:
            b = bytearray(n)
            self._reserved = b
            self._reserved_len = n
            self._reserved_in_active = False
            return memoryview(b)

        if n > self._chunk_size:
            self._flush_active()
            b = bytearray(n)
            self._reserved = b
            self._reserved_len = n
            self._reserved_in_active = False
            return memoryview(b)

        # Ensure reservation fits in active; otherwise flush then create a new one.
        if self._active is not None and (self._active_used + n > self._chunk_size):
            self._flush_active()

        a = self._ensure_active()

        start = self._active_used
        # Reservation does not change _active_used (not readable until commit).
        self._reserved = a
        self._reserved_len = n
        self._reserved_in_active = True
        return memoryview(a)[start:start + n]

    def commit(self, n: int, /) -> None:
        if self._reserved is None:
            raise NoOutstandingReserveByteStreamBufferError('no outstanding reserve')
        if n < 0 or n > self._reserved_len:
            raise ValueError(n)

        if self._reserved_in_active:
            a = self._reserved
            self._reserved = None
            self._reserved_len = 0
            self._reserved_in_active = False

            if self._max_size is not None and self._len + n > self._max_size:
                raise BufferTooLargeByteStreamBufferError('buffer exceeded max_size')

            if n:
                self._active_used += n
                self._len += n

            # Keep active for reuse.
            self._active = a
            return

        b = self._reserved
        self._reserved = None
        self._reserved_len = 0
        self._reserved_in_active = False

        if self._max_size is not None and self._len + n > self._max_size:
            raise BufferTooLargeByteStreamBufferError('buffer exceeded max_size')

        if not n:
            return

        if n == len(b):
            self._segs.append(b)
            self._len += n
        else:
            bb = ByteStreamBuffers.memoryview_to_bytes(memoryview(b)[:n])
            self._segs.append(bb)
            self._len += n

    #

    def advance(self, n: int, /) -> None:
        if n < 0 or n > self._len:
            raise ValueError(n)
        if not n:
            return

        self._len -= n

        while n and self._segs:
            s0 = self._segs[0]

            if s0 is self._active:
                avail0 = self._active_readable_len() - self._head_off
            else:
                avail0 = len(s0) - self._head_off

            if avail0 <= 0:
                popped = self._segs.pop(0)
                if popped is self._active:
                    self._active = None
                    self._active_used = 0
                self._head_off = 0
                continue

            if n < avail0:
                self._head_off += n
                return

            n -= avail0
            popped = self._segs.pop(0)
            if popped is self._active:
                self._active = None
                self._active_used = 0
            self._head_off = 0

        if n:
            raise RuntimeError(n)

    def split_to(self, n: int, /) -> ByteStreamBufferView:
        if n < 0 or n > self._len:
            raise ValueError(n)
        if not n:
            return _EMPTY_DIRECT_BYTE_STREAM_BUFFER_VIEW

        out: ta.List[memoryview] = []
        rem = n

        while rem:
            if not self._segs:
                raise RuntimeError(rem)

            s0 = self._segs[0]

            if s0 is self._active:
                rl = self._active_readable_len()
                if self._head_off >= rl:
                    raise RuntimeError(rem)
                mv0 = memoryview(s0)[self._head_off:rl]
            else:
                mv0 = memoryview(s0)
                if self._head_off:
                    mv0 = mv0[self._head_off:]

            if rem < len(mv0):
                out.append(mv0[:rem])
                self._head_off += rem
                self._len -= n
                return byte_stream_buffer_view_from_segments(out)

            out.append(mv0)
            rem -= len(mv0)
            popped = self._segs.pop(0)
            if popped is self._active:
                self._active = None
                self._active_used = 0
            self._head_off = 0

        self._len -= n
        return byte_stream_buffer_view_from_segments(out)

    def coalesce(self, n: int, /) -> memoryview:
        if n < 0:
            raise ValueError(n)
        if n > self._len:
            raise ValueError(n)
        if not n:
            return memoryview(b'')

        if self._reserved is not None:
            raise OutstandingReserveByteStreamBufferError('outstanding reserve')

        mv0 = self.peek()
        if len(mv0) >= n:
            return mv0[:n]

        out = bytearray(n)
        w = 0

        new_segs: ta.List[ta.Union[bytes, bytearray]] = []

        seg_i = 0
        while w < n and seg_i < len(self._segs):
            s = self._segs[seg_i]
            off = self._head_off if not seg_i else 0

            seg_len = len(s) - off
            if s is self._active and seg_i == (len(self._segs) - 1):
                seg_len = self._active_readable_len() - off

            if seg_len <= 0:
                seg_i += 1
                continue

            take = n - w
            if take > seg_len:
                take = seg_len

            out[w:w + take] = memoryview(s)[off:off + take]
            w += take

            if take < seg_len:
                rem = s[off + take:off + seg_len]
                if rem:
                    new_segs.append(rem)
                seg_i += 1
                break

            seg_i += 1

        if seg_i < len(self._segs):
            new_segs.extend(self._segs[seg_i:])

        self._segs = [bytes(out), *new_segs]
        self._head_off = 0

        self._active = None
        self._active_used = 0

        return memoryview(self._segs[0])[:n]

    def _seg_readable_slice(
            self,
            si: int,
            s: ta.Union[bytes, bytearray],
            last_i: int,
    ) -> ta.Tuple[int, int]:
        """
        Compute the readable offset and length for segment at index si.

        Returns (offset, readable_len) where:
          - offset: byte offset into segment (head_off for si==0, else 0)
          - readable_len: number of readable bytes from offset (0 if segment empty/consumed)

        Handles head offset for first segment and active chunk readable length for last segment.
        """

        off = self._head_off if not si else 0
        seg_len = len(s) - off
        if s is self._active and si == last_i:
            seg_len = self._active_readable_len() - off
        return off, max(0, seg_len)

    def _seg_search_range(
            self,
            start: int,
            limit: int,
            m: int,
            seg_gs: int,
            seg_ge: int,
            seg_len: int,
    ) -> ta.Optional[ta.Tuple[int, int]]:
        """
        Compute local search range within a segment.

        Args:
            start: global start position (user-provided)
            limit: global limit (end - m, last valid position where match can start)
            m: pattern length
            seg_gs: segment global start position
            seg_ge: segment global end position
            seg_len: segment readable length

        Returns (local_start, local_end) if segment overlaps search range, else None.
          - local_start: offset within segment to start searching
          - local_end: offset within segment to end searching (exclusive)
        """

        # Check if segment overlaps search range
        if limit < seg_gs or start >= seg_ge:
            return None

        # Compute local start within segment
        ls = max(0, start - seg_gs)

        # Compute local end: can start match anywhere up to limit, need m bytes
        max_start_in_seg = limit - seg_gs
        end_search = min(max_start_in_seg + m, seg_len)

        # Validate range
        if ls >= end_search:
            return None

        return ls, end_search

    def find(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        start, end = self._norm_slice(start, end)

        m = len(sub)
        if m == 0:
            return start
        if end - start < m:
            return -1

        limit = end - m

        tail = b''
        tail_gstart = 0

        gpos = 0

        last_i = len(self._segs) - 1

        for si, s in enumerate(self._segs):
            off, seg_len = self._seg_readable_slice(si, s, last_i)
            if seg_len <= 0:
                continue

            seg_gs = gpos
            seg_ge = gpos + seg_len

            # Within-segment search
            search_range = self._seg_search_range(start, limit, m, seg_gs, seg_ge, seg_len)
            if search_range is not None:
                ls, end_search = search_range
                idx = s.find(sub, off + ls, off + end_search)
                if idx != -1:
                    return seg_gs + (idx - off)

            if m > 1 and tail:
                head_need = m - 1
                # Only read as many bytes as are actually available in this segment to avoid reading uninitialized data
                # from active chunks.
                head_avail = min(head_need, seg_len)
                if head_avail > 0:
                    head = s[off:off + head_avail]
                    comb = tail + head
                    j = comb.find(sub)
                    if j != -1 and j < len(tail) < j + m:
                        cand = tail_gstart + j
                        if start <= cand <= limit:
                            return cand

            if m > 1:
                take = m - 1
                if seg_len >= take:
                    tail = s[off + seg_len - take:off + seg_len]
                    tail_gstart = seg_ge - take
                else:
                    tail = (tail + s[off:off + seg_len])[-(m - 1):]
                    tail_gstart = seg_ge - len(tail)

            gpos = seg_ge

        return -1

    def rfind(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        start, end = self._norm_slice(start, end)

        m = len(sub)
        if m == 0:
            return end
        if end - start < m:
            return -1

        limit = end - m

        if not self._segs:
            return -1

        best = -1

        seg_ge = self._len
        prev_s: ta.Optional[ta.Union[bytes, bytearray]] = None
        prev_off = 0
        prev_seg_len = 0

        last_i = len(self._segs) - 1

        for si in range(len(self._segs) - 1, -1, -1):
            s = self._segs[si]
            off, seg_len = self._seg_readable_slice(si, s, last_i)
            if seg_len <= 0:
                continue

            seg_gs = seg_ge - seg_len

            # Within-segment search
            search_range = self._seg_search_range(start, limit, m, seg_gs, seg_ge, seg_len)
            if search_range is not None:
                ls, end_search = search_range
                idx = s.rfind(sub, off + ls, off + end_search)
                if idx != -1:
                    cand = seg_gs + (idx - off)
                    if cand > best:
                        best = cand

            if m > 1 and prev_s is not None:
                tail_need = m - 1
                if seg_len >= tail_need:
                    tail = s[off + seg_len - tail_need:off + seg_len]
                    tail_gstart = seg_ge - tail_need

                else:
                    tail_parts = [s[off:off + seg_len]]
                    tail_len = seg_len
                    for sj in range(si - 1, -1, -1):
                        if tail_len >= tail_need:
                            break

                        sj_s = self._segs[sj]
                        sj_off, sj_len = self._seg_readable_slice(sj, sj_s, last_i)
                        if sj_len <= 0:
                            continue

                        take = min(tail_need - tail_len, sj_len)
                        tail_parts.insert(0, sj_s[sj_off + sj_len - take:sj_off + sj_len])
                        tail_len += take

                    tail_combined = b''.join(tail_parts)
                    tail = tail_combined[-(m - 1):] if len(tail_combined) >= m - 1 else tail_combined
                    tail_gstart = seg_ge - len(tail)

                head_need = m - 1
                # Only read as many bytes as are actually available in prev segment to avoid reading uninitialized data
                # from active chunks.
                head_avail = min(head_need, prev_seg_len)
                if head_avail > 0:
                    head = prev_s[prev_off:prev_off + head_avail]
                else:
                    head = b''

                comb = tail + head
                j = comb.rfind(sub)
                if j != -1 and j < len(tail) < j + m:
                    cand = tail_gstart + j
                    if start <= cand <= limit and cand > best:
                        best = cand

            if best >= seg_gs:
                return best

            prev_s = s
            prev_off = off
            prev_seg_len = seg_len
            seg_ge = seg_gs

        return best


##


def byte_stream_buffer_view_from_segments(mvs: ta.Sequence[memoryview]) -> ByteStreamBufferView:
    if not mvs:
        return _EMPTY_DIRECT_BYTE_STREAM_BUFFER_VIEW
    elif len(mvs) == 1:
        return DirectByteStreamBufferView(mvs[0])
    else:
        return SegmentedByteStreamBufferView(mvs)



import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .states import LifecycleState


##


class Lifecycle(lang.Abstract):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        check.not_issubclass_except_nameerror(cls, lambda: AsyncLifecycle)

    def lifecycle_state(self, state: LifecycleState) -> None:
        """
        This method is not intended for actual work - this is intended solely for safe bookkeeping. Only do real work in
        the explicit per-state methods.
        """

    def lifecycle_construct(self) -> None:
        pass

    def lifecycle_start(self) -> None:
        pass

    def lifecycle_stop(self) -> None:
        pass

    def lifecycle_destroy(self) -> None:
        pass


class AsyncLifecycle(lang.Abstract):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        check.not_issubclass(cls, Lifecycle)

    async def lifecycle_state(self, state: LifecycleState) -> None:
        """
        This method is not intended for actual work - this is intended solely for safe bookkeeping. Only do real work in
        the explicit per-state methods.
        """

    async def lifecycle_construct(self) -> None:
        pass

    async def lifecycle_start(self) -> None:
        pass

    async def lifecycle_stop(self) -> None:
        pass

    async def lifecycle_destroy(self) -> None:
        pass


AnyLifecycle: ta.TypeAlias = Lifecycle | AsyncLifecycle

ANY_LIFECYCLE_TYPES: tuple[type[Lifecycle | AsyncLifecycle], ...] = (Lifecycle, AsyncLifecycle)


##


@ta.final
@dc.dataclass(frozen=True)
class _SyncToAsyncLifecycle(AsyncLifecycle, lang.Final):
    lifecycle: Lifecycle

    async def lifecycle_state(self, state: LifecycleState) -> None:
        self.lifecycle.lifecycle_state(state)

    async def lifecycle_construct(self) -> None:
        self.lifecycle.lifecycle_construct()

    async def lifecycle_start(self) -> None:
        self.lifecycle.lifecycle_start()

    async def lifecycle_stop(self) -> None:
        self.lifecycle.lifecycle_stop()

    async def lifecycle_destroy(self) -> None:
        self.lifecycle.lifecycle_destroy()


def sync_to_async_lifecycle(lifecycle: Lifecycle) -> AsyncLifecycle:
    return _SyncToAsyncLifecycle(lifecycle)


def as_async_lifecycle(lifecycle: Lifecycle | AsyncLifecycle) -> AsyncLifecycle:
    if isinstance(lifecycle, Lifecycle):
        return sync_to_async_lifecycle(lifecycle)

    elif isinstance(lifecycle, AsyncLifecycle):
        return lifecycle

    else:
        raise TypeError(lifecycle)


@ta.final
@dc.dataclass(frozen=True)
class _AsyncToSyncLifecycle(Lifecycle, lang.Final):
    lifecycle: AsyncLifecycle

    def lifecycle_state(self, state: LifecycleState) -> None:
        lang.sync_await(self.lifecycle.lifecycle_state(state))

    def lifecycle_construct(self) -> None:
        lang.sync_await(self.lifecycle.lifecycle_construct())

    def lifecycle_start(self) -> None:
        lang.sync_await(self.lifecycle.lifecycle_start())

    def lifecycle_stop(self) -> None:
        lang.sync_await(self.lifecycle.lifecycle_stop())

    def lifecycle_destroy(self) -> None:
        lang.sync_await(self.lifecycle.lifecycle_destroy())


def async_to_sync_lifecycle(lifecycle: AsyncLifecycle) -> Lifecycle:
    return _AsyncToSyncLifecycle(lifecycle)


def as_sync_lifecycle(lifecycle: Lifecycle | AsyncLifecycle) -> Lifecycle:
    if isinstance(lifecycle, Lifecycle):
        return lifecycle

    elif isinstance(lifecycle, AsyncLifecycle):
        return async_to_sync_lifecycle(lifecycle)

    else:
        raise TypeError(lifecycle)


##


@ta.final
@dc.dataclass(frozen=True, kw_only=True)
class CallbackLifecycle(Lifecycle, lang.Final):
    on_state: ta.Callable[[LifecycleState], None] | None = None
    on_construct: ta.Callable[[], None] | None = None
    on_start: ta.Callable[[], None] | None = None
    on_stop: ta.Callable[[], None] | None = None
    on_destroy: ta.Callable[[], None] | None = None

    @ta.override
    def lifecycle_state(self, state: LifecycleState) -> None:
        if self.on_state is not None:
            self.on_state(state)

    @ta.override
    def lifecycle_construct(self) -> None:
        if self.on_construct is not None:
            self.on_construct()

    @ta.override
    def lifecycle_start(self) -> None:
        if self.on_start is not None:
            self.on_start()

    @ta.override
    def lifecycle_stop(self) -> None:
        if self.on_stop is not None:
            self.on_stop()

    @ta.override
    def lifecycle_destroy(self) -> None:
        if self.on_destroy is not None:
            self.on_destroy()


@ta.final
@dc.dataclass(frozen=True, kw_only=True)
class CallbackAsyncLifecycle(AsyncLifecycle, lang.Final):
    on_state: ta.Callable[[LifecycleState], ta.Awaitable[None]] | None = None
    on_construct: ta.Callable[[], ta.Awaitable[None]] | None = None
    on_start: ta.Callable[[], ta.Awaitable[None]] | None = None
    on_stop: ta.Callable[[], ta.Awaitable[None]] | None = None
    on_destroy: ta.Callable[[], ta.Awaitable[None]] | None = None

    @ta.override
    async def lifecycle_state(self, state: LifecycleState) -> None:
        if self.on_state is not None:
            await self.on_state(state)

    @ta.override
    async def lifecycle_construct(self) -> None:
        if self.on_construct is not None:
            await self.on_construct()

    @ta.override
    async def lifecycle_start(self) -> None:
        if self.on_start is not None:
            await self.on_start()

    @ta.override
    async def lifecycle_stop(self) -> None:
        if self.on_stop is not None:
            await self.on_stop()

    @ta.override
    async def lifecycle_destroy(self) -> None:
        if self.on_destroy is not None:
            await self.on_destroy()



# ruff: noqa: UP045
# @omlish-lite
import threading
import typing as ta

from ..locks import AsyncliteLock
from ..locks import AsyncliteLocks
from .base import SyncAsyncliteApi
from .base import SyncAsyncliteObject


##


class SyncAsyncliteLock(AsyncliteLock, SyncAsyncliteObject):
    def __init__(self, u: threading.Lock) -> None:
        super().__init__()

        self._u = u

    async def acquire(self, *, timeout: ta.Optional[float] = None) -> None:
        if timeout is not None and timeout > 0:
            a = self._u.acquire(blocking=False, timeout=timeout)
        else:
            a = self._u.acquire(blocking=True)

        if not a:
            raise TimeoutError

    def acquire_nowait(self) -> bool:
        return self._u.acquire(blocking=False)

    def release(self) -> None:
        self._u.release()

    def locked(self) -> bool:
        return self._u.locked()


class SyncAsyncliteLocks(AsyncliteLocks, SyncAsyncliteApi):
    def make_lock(self) -> AsyncliteLock:
        return SyncAsyncliteLock(threading.Lock())



# ruff: noqa: UP045
# @omlish-lite
import abc
import contextlib
import dataclasses as dc
import typing as ta

from ...io.readers import BufferedBytesReader
from ...lite.abstract import Abstract
from ...lite.dataclasses import dataclass_shallow_asdict
from .base import BaseHttpClient
from .base import BaseHttpResponse
from .base import BaseHttpResponseT
from .base import HttpClientContext
from .base import HttpRequest
from .base import HttpResponse
from .base import HttpStatusError


StreamHttpResponseT = ta.TypeVar('StreamHttpResponseT', bound='StreamHttpResponse')
HttpClientT = ta.TypeVar('HttpClientT', bound='HttpClient')


##


@ta.final
@dc.dataclass(frozen=True)  # kw_only=True
class StreamHttpResponse(BaseHttpResponse):
    _stream: ta.Optional[BufferedBytesReader] = None

    @property
    def stream(self) -> 'BufferedBytesReader':
        if (st := self._stream) is None:
            raise TypeError('No data')
        return st

    @property
    def has_data(self) -> bool:
        return self._stream is not None

    #

    _closer: ta.Optional[ta.Callable[[], None]] = None

    def __enter__(self: StreamHttpResponseT) -> StreamHttpResponseT:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self) -> None:
        if (c := self._closer) is not None:
            c()  # noqa


#


def close_http_client_response(resp: BaseHttpResponse) -> None:
    if isinstance(resp, HttpResponse):
        pass

    elif isinstance(resp, StreamHttpResponse):
        resp.close()

    else:
        raise TypeError(resp)


@contextlib.contextmanager
def closing_http_client_response(resp: BaseHttpResponseT) -> ta.Iterator[BaseHttpResponseT]:
    if isinstance(resp, HttpResponse):
        yield resp
        return

    elif isinstance(resp, StreamHttpResponse):
        with contextlib.closing(resp):
            yield resp

    else:
        raise TypeError(resp)


def read_http_client_response(resp: BaseHttpResponse) -> HttpResponse:
    if isinstance(resp, HttpResponse):
        return resp

    elif isinstance(resp, StreamHttpResponse):
        return HttpResponse(**{
            **{k: v for k, v in dataclass_shallow_asdict(resp).items() if k not in ('_stream', '_closer')},
            **({'data': resp.stream.readall()} if resp.has_data else {}),
        })

    else:
        raise TypeError(resp)


##


class HttpClient(BaseHttpClient, Abstract):
    def __enter__(self: HttpClientT) -> HttpClientT:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def request(
            self,
            req: HttpRequest,
            *,
            context: ta.Optional[HttpClientContext] = None,
            check: bool = False,
    ) -> HttpResponse:
        with closing_http_client_response(self.stream_request(
                req,
                context=context,
                check=check,
        )) as resp:
            return read_http_client_response(resp)

    def stream_request(
            self,
            req: HttpRequest,
            *,
            context: ta.Optional[HttpClientContext] = None,
            check: bool = False,
    ) -> StreamHttpResponse:
        if context is None:
            context = HttpClientContext()

        resp = self._stream_request(context, req)

        try:
            if check and not resp.is_success:
                if isinstance(resp.underlying, Exception):
                    cause = resp.underlying
                else:
                    cause = None
                raise HttpStatusError(read_http_client_response(resp)) from cause  # noqa

        except Exception:
            close_http_client_response(resp)
            raise

        return resp

    @abc.abstractmethod
    def _stream_request(self, ctx: HttpClientContext, req: HttpRequest) -> StreamHttpResponse:
        raise NotImplementedError



import dataclasses as dc
import heapq
import itertools
import time

from ... import check
from .base import BaseScheduledRunner
from .handles import ScheduleHandleImpl
from .types import ScheduledFn
from .types import ScheduledRunnerState
from .types import ScheduledRunnerStateError
from .types import ScheduleHandle


##


class HeapqScheduledRunner(BaseScheduledRunner):
    """
    Single-thread scheduled executor.

    - Uses time.monotonic()
    - Executes fn in the runner thread (keep it non-blocking)
    - Thread-safe
    """

    def __init__(
            self,
            *,
            batch_size: int | None = None,
            batch_max_time_s: float | None = None,
    ) -> None:
        super().__init__()

        if batch_size is None:
            batch_size = 1
        if not (batch_size > 0):
            raise ValueError('batch_size must be positive')
        self._batch_size = batch_size
        self._batch_max_time_s = batch_max_time_s

        self._heap: list[tuple[float, int, HeapqScheduledRunner._Task]] = []

    _seq = 0

    @dc.dataclass()
    class _Task:
        when: float
        seq: int
        fn: ScheduledFn
        handle: 'ScheduleHandleImpl'

        running: bool = False

    def schedule(self, delay: float, fn: ScheduledFn) -> ScheduleHandle:
        when = time.monotonic() + max(0., delay)

        with self._cv:
            self._ensure_thread_locked()
            self._check_state(ScheduledRunnerState.RUNNING)

            self._seq += 1

            h = ScheduleHandleImpl(self._cv)

            task = HeapqScheduledRunner._Task(
                when=when,
                seq=self._seq,
                fn=fn,
                handle=h,
            )

            heapq.heappush(self._heap, (task.when, task.seq, task))

            # Wake if this is the earliest task
            if self._heap[0][2] is task:
                self._cv.notify()

            return h

    def _thread_main(self) -> None:
        recycle: list[HeapqScheduledRunner._Task] | None = None
        cancel: list[ScheduleHandle] | None = None

        for epoch in itertools.count():  # noqa
            popped: HeapqScheduledRunner._Task | list[HeapqScheduledRunner._Task] | None = None

            with self._cv:
                # Recycle unhandled popped tasks from last iteration

                if recycle:
                    for t in recycle:
                        check.state(not t.running)
                        heapq.heappush(self._heap, (t.when, t.seq, t))
                    recycle = None

                # Skip cancelled/done tasks at heap head

                while self._heap:
                    _when, _seq, t = self._heap[0]
                    if not t.handle.done():
                        break
                    heapq.heappop(self._heap)

                # Check state

                if self._state == ScheduledRunnerState.RUNNING:
                    pass

                elif self._state == ScheduledRunnerState.STOPPING:
                    cancel = []
                    for _, _, t in self._heap:
                        check.state(not t.running)
                        cancel.append(t.handle)
                    self._heap.clear()
                    break

                elif self._state == ScheduledRunnerState.DRAINING:
                    if not self._heap:
                        break

                else:
                    raise ScheduledRunnerStateError(self._state)

                # Wait for runnable heap

                if not self._heap:
                    self._cv.wait()
                    continue

                now = time.monotonic()

                when, _seq, t = self._heap[0]
                check.state(not t.running)
                if when > now:
                    self._cv.wait(timeout=(when - now))
                    continue

                # Pop task chunk

                n = 0
                while n < self._batch_size and self._heap:
                    when, _seq, t = self._heap[0]
                    check.state(not t.running)
                    if when > now:
                        break

                    heapq.heappop(self._heap)
                    if t.handle.done():
                        continue

                    if not n:
                        popped = t
                    elif n == 1:
                        popped = [popped, t]  # type: ignore
                    else:
                        popped.append(t)  # type: ignore

                    n += 1

            # Run popped tasks

            if n == 1:
                self._run_task(popped)  # type: ignore

            elif n > 1:
                if (bmt := self._batch_max_time_s) is not None:
                    deadline: float | None = now + bmt
                else:
                    deadline = None

                i = 0
                for t in popped:  # type: ignore
                    i += 1
                    self._run_task(t)
                    if deadline is not None and time.monotonic() >= deadline:
                        break

                if i < len(popped):  # type: ignore
                    recycle = popped[i:]  # type: ignore

        if cancel:
            for h in cancel:
                h.cancel()

    def _run_task(self, task: _Task) -> None:
        check.state(not task.running)
        task.running = True
        task.handle._set_running()  # noqa

        # execute without lock
        if task.handle.cancelled():
            return

        try:
            res = task.fn()
        except BaseException as e:  # noqa
            task.handle._set_exception(e)  # noqa
        else:
            task.handle._set_result(res)  # noqa
        finally:
            task.running = False



from ... import dataclasses as _dc  # noqa


_dc.init_package(
    globals(),
    codegen=True,
)


##


from .base import (  # noqa
    Op,
)

from .core import (  # noqa
    CORE_RULES,
)

from .errors import (  # noqa
    AbnfError,
    AbnfGrammarParseError,
)

from .grammars import (  # noqa
    Channel,
    Rule,
    RulesCollection,

    Grammar,
)

from .matches import (  # noqa
    Match,

    longest_match,
    furthest_match,
    filter_matches,
)

from .meta import (  # noqa
    META_GRAMMAR_RULES,
    META_GRAMMAR,

    parse_grammar,
)

from .ops import (  # noqa
    Literal,
    StringLiteral,
    CaseInsensitiveStringLiteral,
    RangeLiteral,
    literal,

    Concat,
    concat,

    Repeat,
    repeat,
    option,

    Either,
    either,

    RuleRef,
    rule,
)

from .parsing import (  # noqa
    iter_parse,
    parse,
)

from .utils import (  # noqa
    filter_match_channels,

    only_match_rules,

    fix_ws,
)
